@@ -28,7 +28,6 @@
 #include <stdio.h>
 #include <assert.h>
 #include "system.h"
-#include "closeout.h"
 #include "long-options.h"
 #include "error.h"
 #include "hard-locale.h"
@@ -224,11 +223,20 @@ static MONTHTAB_CONST struct month monthtab[] =
 /* During the merge phase, the number of files to merge at once. */
 #define NMERGE 16
 
-/* The approximate maximum number of bytes of main memory to use.  */
-static size_t sort_size;
+/* Minimum size for a merge or check buffer.  */
+#define MIN_MERGE_BUFFER_SIZE (2 + sizeof (struct line))
 
 /* Minimum sort size; the code might not work with smaller sizes.  */
-#define MIN_SORT_SIZE (NMERGE * (2 + sizeof (struct line)))
+#define MIN_SORT_SIZE (NMERGE * MIN_MERGE_BUFFER_SIZE)
+
+/* The number of bytes needed for a merge or check buffer, which can
+   function relatively efficiently even if it holds only one line.  If
+   a longer line is seen, this value is increased.  */
+static size_t merge_buffer_size = MAX (MIN_MERGE_BUFFER_SIZE, 256 * 1024);
+
+/* The approximate maximum number of bytes of main memory to use, as
+   specified by the user.  Zero if the user has not specified a size.  */
+static size_t sort_size;
 
 /* The guessed size for non-regular files.  */
 #define INPUT_FILE_SIZE_GUESS (1024 * 1024)
@@ -283,18 +291,13 @@ Usage: %s [OPTION]... [FILE]...\n\
       printf (_("\
 Write sorted concatenation of all FILE(s) to standard output.\n\
 \n\
-  +POS1 [-POS2]    start a key at POS1, end it *before* POS2 (obsolescent)\n\
-                     field numbers and character offsets are numbered\n\
-                     starting with zero (contrast with the -k option)\n\
   -b               ignore leading blanks in sort fields or keys\n\
   -c               check if given files already sorted, do not sort\n\
   -d               consider only blanks and alphanumeric characters in keys\n\
   -f               fold lower case to upper case characters in keys\n\
   -g               compare according to general numerical value, imply -b\n\
   -i               consider only printable characters in keys\n\
-  -k POS1[,POS2]   start a key at POS1, end it *at* POS2\n\
-                     field numbers and character offsets are numbered\n\
-                     starting with one (contrast with zero-based +POS form)\n\
+  -k POS1[,POS2]   start a key at POS1, end it *at* POS2 (origin 1)\n\
   -m               merge already sorted files, do not sort\n\
   -M               compare (unknown) < `JAN' < ... < `DEC', imply -b\n\
   -n               compare according to string numerical value, imply -b\n\
@@ -308,6 +311,9 @@ Write sorted concatenation of all FILE(s) to standard output.\n\
   -u               with -c, check for strict ordering;\n\
                    with -m, only output the first of an equal sequence\n\
   -z               end lines with 0 byte, not newline, for find -print0\n\
+  +POS1 [-POS2]    start a key at POS1, end it *before* POS2 (origin 0)\n\
+                     Warning: this option is obsolescent and support for it\n\
+                     will be withdrawn.  Use -k instead.\n\
       --help       display this help and exit\n\
       --version    output version information and exit\n\
 \n\
@@ -361,13 +367,13 @@ cleanup (void)
     unlink (node->name);
 }
 
-/* Report an error for STRING, clean up, and exit.  */
+/* Report MESSAGE for FILE, then clean up and exit.  */
 
-static void die PARAMS ((char const *)) ATTRIBUTE_NORETURN;
+static void die PARAMS ((char const *, char const *)) ATTRIBUTE_NORETURN;
 static void
-die (char const *string)
+die (char const *message, char const *file)
 {
-  error (0, errno, "%s", string);
+  error (0, errno, "%s: %s", message, file);
   cleanup ();
   exit (SORT_FAILURE);
 }
@@ -405,7 +411,7 @@ create_temp_file (FILE **pfp)
   errno = saved_errno;
 
   if (fd < 0 || (*pfp = fdopen (fd, "w")) == NULL)
-    die (file);
+    die (_("cannot create temporary file"), file);
 
   return file;
 }
@@ -417,13 +423,18 @@ xfopen (const char *file, const char *how)
 
   if (STREQ (file, "-"))
     {
-      have_read_stdin = 1;
-      fp = stdin;
+      if (*how == 'r')
+	{
+	  have_read_stdin = 1;
+	  fp = stdin;
+	}
+      else
+	fp = stdout;
     }
   else
     {
       if ((fp = fopen_safer (file, how)) == NULL)
-	die (file);
+	die (_("open failed"), file);
     }
 
   return fp;
@@ -443,7 +454,7 @@ xfclose (FILE *fp, char const *file)
   else
     {
       if (fclose (fp) != 0)
-	die (file);
+	die (_("close failed"), file);
     }
 }
 
@@ -451,7 +462,7 @@ static void
 write_bytes (const char *buf, size_t n_bytes, FILE *fp, const char *output_file)
 {
   if (fwrite (buf, 1, n_bytes, fp) != n_bytes)
-    die (output_file);
+    die (_("write failed"), output_file);
 }
 
 /* Append DIR to the array of temporary directory names.  */
@@ -658,7 +669,7 @@ sort_buffer_size (FILE *const *fps, int nfps,
 	   : strcmp (files[i], "-") == 0 ? fstat (STDIN_FILENO, &st)
 	   : stat (files[i], &st))
 	  != 0)
-	die (files[i]);
+	die (_("stat failed"), files[i]);
 
       file_size = S_ISREG (st.st_mode) ? st.st_size : INPUT_FILE_SIZE_GUESS;
 
@@ -873,6 +884,7 @@ fillbuf (struct buffer *buf, register FILE *fp, char const *file)
   struct keyfield const *key = keylist;
   char eol = eolchar;
   size_t line_bytes = buf->line_bytes;
+  size_t mergesize = merge_buffer_size - MIN_MERGE_BUFFER_SIZE;
 
   if (buf->eof)
     return 0;
@@ -908,7 +920,7 @@ fillbuf (struct buffer *buf, register FILE *fp, char const *file)
 	  if (bytes_read != readsize)
 	    {
 	      if (ferror (fp))
-		die (file);
+		die (_("read failed"), file);
 	      if (feof (fp))
 		{
 		  buf->eof = 1;
@@ -926,6 +938,7 @@ fillbuf (struct buffer *buf, register FILE *fp, char const *file)
 	      line--;
 	      line->text = line_start;
 	      line->length = ptr - line_start;
+	      mergesize = MAX (mergesize, line->length);
 	      avail -= line_bytes;
 
 	      if (key)
@@ -960,6 +973,7 @@ fillbuf (struct buffer *buf, register FILE *fp, char const *file)
       if (buf->nlines != 0)
 	{
 	  buf->left = ptr - line_start;
+	  merge_buffer_size = mergesize + MIN_MERGE_BUFFER_SIZE;
 	  return 1;
 	}
 
@@ -1488,8 +1502,7 @@ checkfp (FILE *fp, char *file_name)
   int disordered = 0;
 
   initbuf (&buf, sizeof (struct line),
-	   sort_buffer_size (&fp, 1, &file_name, 1,
-			     sizeof (struct line), sort_size));
+	   MAX (merge_buffer_size, sort_size));
   temp.text = NULL;
 
   while (fillbuf (&buf, fp, file_name))
@@ -1556,14 +1569,16 @@ checkfp (FILE *fp, char *file_name)
   return disordered;
 }
 
-/* Merge lines from FPS onto OFP.  NFPS cannot be greater than NMERGE.
-   Close FPS before returning.  NAMES and OUTPUT_NAME give the names
-   of the files.  */
+/* Merge lines from FILES onto OFP.  NFILES cannot be greater than
+   NMERGE.  Close input and output files before returning.
+   OUTPUT_FILE gives the name of the output file; if OFP is NULL, the
+   output file has not been opened yet.  */
 
 static void
-mergefps (FILE **fps, char **files, register int nfps,
+mergefps (char **files, register int nfiles,
 	  FILE *ofp, const char *output_file)
 {
+  FILE *fps[NMERGE];		/* Input streams for each file.  */
   struct buffer buffer[NMERGE];	/* Input buffers for each file. */
   struct line saved;		/* Saved line storage for unique check. */
   struct line const *savedline = NULL;
@@ -1579,25 +1594,21 @@ mergefps (FILE **fps, char **files, register int nfps,
   saved.text = NULL;
 
   /* Read initial lines from each input file. */
-  for (i = 0; i < nfps; ++i)
+  for (i = 0; i < nfiles; ++i)
     {
-      size_t mergealloc = sort_buffer_size (&fps[i], 1, &files[i], 1,
-					    sizeof (struct line),
-					    sort_size / nfps);
-      initbuf (&buffer[i], sizeof (struct line), mergealloc);
+      fps[i] = xfopen (files[i], "r");
+      initbuf (&buffer[i], sizeof (struct line),
+	       MAX (merge_buffer_size, sort_size));
       /* If a file is empty, eliminate it from future consideration. */
-      while (i < nfps && !fillbuf (&buffer[i], fps[i], files[i]))
+      while (i < nfiles && !fillbuf (&buffer[i], fps[i], files[i]))
 	{
 	  xfclose (fps[i], files[i]);
 	  zaptemp (files[i]);
-	  --nfps;
-	  for (j = i; j < nfps; ++j)
-	    {
-	      fps[j] = fps[j + 1];
-	      files[j] = files[j + 1];
-	    }
+	  --nfiles;
+	  for (j = i; j < nfiles; ++j)
+	    files[j] = files[j + 1];
 	}
-      if (i == nfps)
+      if (i == nfiles)
 	free (buffer[i].buf);
       else
 	{
@@ -1607,17 +1618,20 @@ mergefps (FILE **fps, char **files, register int nfps,
 	}
     }
 
+  if (! ofp)
+    ofp = xfopen (output_file, "w");
+
   /* Set up the ord table according to comparisons among input lines.
      Since this only reorders two items if one is strictly greater than
      the other, it is stable. */
-  for (i = 0; i < nfps; ++i)
+  for (i = 0; i < nfiles; ++i)
     ord[i] = i;
-  for (i = 1; i < nfps; ++i)
+  for (i = 1; i < nfiles; ++i)
     if (0 < compare (cur[ord[i - 1]], cur[ord[i]]))
       t = ord[i - 1], ord[i - 1] = ord[i], ord[i] = t, i = 0;
 
   /* Repeatedly output the smallest line until no input remains. */
-  while (nfps)
+  while (nfiles)
     {
       struct line const *smallest = cur[ord[0]];
 
@@ -1673,14 +1687,14 @@ mergefps (FILE **fps, char **files, register int nfps,
 	  else
 	    {
 	      /* We reached EOF on fps[ord[0]]. */
-	      for (i = 1; i < nfps; ++i)
+	      for (i = 1; i < nfiles; ++i)
 		if (ord[i] > ord[0])
 		  --ord[i];
-	      --nfps;
+	      --nfiles;
 	      xfclose (fps[ord[0]], files[ord[0]]);
 	      zaptemp (files[ord[0]]);
 	      free (buffer[ord[0]].buf);
-	      for (i = ord[0]; i < nfps; ++i)
+	      for (i = ord[0]; i < nfiles; ++i)
 		{
 		  fps[i] = fps[i + 1];
 		  files[i] = files[i + 1];
@@ -1688,7 +1702,7 @@ mergefps (FILE **fps, char **files, register int nfps,
 		  cur[i] = cur[i + 1];
 		  base[i] = base[i + 1];
 		}
-	      for (i = 0; i < nfps; ++i)
+	      for (i = 0; i < nfiles; ++i)
 		ord[i] = ord[i + 1];
 	      continue;
 	    }
@@ -1697,7 +1711,7 @@ mergefps (FILE **fps, char **files, register int nfps,
       /* The new line just read in may be larger than other lines
 	 already in core; push it back in the queue until we encounter
 	 a line larger than it. */
-      for (i = 1; i < nfps; ++i)
+      for (i = 1; i < nfiles; ++i)
 	{
 	  t = compare (cur[ord[0]], cur[ord[i]]);
 	  if (!t)
@@ -1716,6 +1730,8 @@ mergefps (FILE **fps, char **files, register int nfps,
       write_bytes (saved.text, saved.length, ofp, output_file);
       free (saved.text);
     }
+
+  xfclose (ofp, output_file);
 }
 
 /* Sort the array LINES with NLINES members, using TEMP for temporary space.
@@ -1764,6 +1780,46 @@ sortlines (struct line *lines, size_t nlines, struct line *temp)
     *--lo = *--t;
 }
 
+/* Return the index of the first of NFILES FILES that is the same file
+   as OUTFILE.  If none can be the same, return NFILES.  Consider an
+   input pipe to be the same as OUTFILE, since the pipe might be the
+   output of a command like "cat OUTFILE".  */
+
+static int
+first_same_file (char **files, int nfiles, char const *outfile)
+{
+  int i;
+  int got_outstat = 0;
+  struct stat instat, outstat;
+
+  for (i = 0; i < nfiles; i++)
+    {
+      int standard_input = STREQ (files[i], "-");
+
+      if (STREQ (outfile, files[i]) && ! standard_input)
+	return i;
+
+      if (! got_outstat)
+	{
+	  got_outstat = 1;
+	  if ((STREQ (outfile, "-")
+	       ? fstat (STDOUT_FILENO, &outstat)
+	       : stat (outfile, &outstat))
+	      != 0)
+	    return nfiles;
+	}
+
+      if (((standard_input
+	    ? fstat (STDIN_FILENO, &instat)
+	    : stat (files[i], &instat))
+	   == 0)
+	  && (S_ISFIFO (instat.st_mode) || SAME_INODE (instat, outstat)))
+	return i;
+    }
+
+  return nfiles;
+}
+
 /* Check that each of the NFILES FILES is ordered.
    Return a count of disordered files. */
 
@@ -1781,51 +1837,47 @@ check (char **files, int nfiles)
   return disorders;
 }
 
-/* Merge NFILES FILES onto OFP. */
+/* Merge NFILES FILES onto OUTPUT_FILE.  However, merge at most
+   MAX_MERGE input files directly onto OUTPUT_FILE.  MAX_MERGE cannot
+   exceed NMERGE.  */
 
 static void
-merge (char **files, int nfiles, FILE *ofp, const char *output_file)
+merge (char **files, int nfiles, int max_merge, char const *output_file)
 {
-  int i, j, t;
-  char *temp;
-  FILE *fps[NMERGE], *tfp;
-
-  while (nfiles > NMERGE)
+  while (max_merge < nfiles)
     {
-      t = 0;
+      FILE *tfp;
+      int i, t = 0;
+      char *temp;
       for (i = 0; i < nfiles / NMERGE; ++i)
 	{
-	  for (j = 0; j < NMERGE; ++j)
-	    fps[j] = xfopen (files[i * NMERGE + j], "r");
 	  temp = create_temp_file (&tfp);
-	  mergefps (fps, &files[i * NMERGE], NMERGE, tfp, temp);
-	  xfclose (tfp, temp);
+	  mergefps (&files[i * NMERGE], NMERGE, tfp, temp);
 	  files[t++] = temp;
 	}
-      for (j = 0; j < nfiles % NMERGE; ++j)
-	fps[j] = xfopen (files[i * NMERGE + j], "r");
       temp = create_temp_file (&tfp);
-      mergefps (fps, &files[i * NMERGE], nfiles % NMERGE, tfp, temp);
-      xfclose (tfp, temp);
+      mergefps (&files[i * NMERGE], nfiles % NMERGE, tfp, temp);
       files[t++] = temp;
       nfiles = t;
+      if (nfiles == 1)
+	break;
     }
 
-  for (i = 0; i < nfiles; ++i)
-    fps[i] = xfopen (files[i], "r");
-  mergefps (fps, files, nfiles, ofp, output_file);
+  mergefps (files, nfiles, NULL, output_file);
 }
 
-/* Sort NFILES FILES onto OFP. */
+/* Sort NFILES FILES onto OUTPUT_FILE. */
 
 static void
-sort (char **files, int nfiles, FILE *ofp, const char *output_file)
+sort (char **files, int nfiles, char const *output_file)
 {
   struct buffer buf;
-  FILE *tfp;
-  struct tempnode *node;
   int n_temp_files = 0;
-  char **tempfiles;
+  int output_file_created = 0;
+
+  static size_t size;
+  if (! size && ! (size = sort_size))
+    size = default_sort_size ();
 
   buf.alloc = 0;
 
@@ -1834,11 +1886,12 @@ sort (char **files, int nfiles, FILE *ofp, const char *output_file)
       char const *temp_output;
       char const *file = *files;
       FILE *fp = xfopen (file, "r");
+      FILE *tfp;
 
       if (! buf.alloc)
 	initbuf (&buf, 2 * sizeof (struct line),
 		 sort_buffer_size (&fp, 1, files, nfiles,
-				   2 * sizeof (struct line), sort_size));
+				   2 * sizeof (struct line), size));
       buf.eof = 0;
       files++;
       nfiles--;
@@ -1865,8 +1918,10 @@ sort (char **files, int nfiles, FILE *ofp, const char *output_file)
 	  sortlines (line, buf.nlines, linebase);
 	  if (buf.eof && !nfiles && !n_temp_files && !buf.left)
 	    {
-	      tfp = ofp;
+	      xfclose (fp, file);
+	      tfp = xfopen (output_file, "w");
 	      temp_output = output_file;
+	      output_file_created = 1;
 	    }
 	  else
 	    {
@@ -1884,21 +1939,25 @@ sort (char **files, int nfiles, FILE *ofp, const char *output_file)
 	    }
 	  while (linebase < line);
 
-	  if (tfp != ofp)
-	    xfclose (tfp, temp_output);
+	  xfclose (tfp, temp_output);
+
+	  if (output_file_created)
+	    goto finish;
 	}
       xfclose (fp, file);
     }
 
+ finish:
   free (buf.buf);
 
-  if (n_temp_files)
+  if (! output_file_created)
     {
       int i = n_temp_files;
-      tempfiles = (char **) xmalloc (n_temp_files * sizeof (char *));
+      struct tempnode *node;
+      char **tempfiles = (char **) xmalloc (n_temp_files * sizeof (char *));
       for (node = temphead; i > 0; node = node->next)
 	tempfiles[--i] = node->name;
-      merge (tempfiles, n_temp_files, ofp, output_file);
+      merge (tempfiles, n_temp_files, NMERGE, output_file);
       free ((char *) tempfiles);
     }
 }
@@ -2044,9 +2103,8 @@ main (int argc, char **argv)
   char const *s;
   int i;
   int checkonly = 0, mergeonly = 0, nfiles = 0;
-  char *minus = "-", **files, *tmp;
+  char *minus = "-", **files;
   char const *outfile = minus;
-  FILE *ofp;
   static int const sigs[] = { SIGHUP, SIGINT, SIGPIPE, SIGTERM };
   int nsigs = sizeof sigs / sizeof *sigs;
 #ifdef SA_NOCLDSTOP
@@ -2059,9 +2117,6 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  close_stdout_set_status (SORT_FAILURE);
-  atexit (close_stdout);
-
 #ifdef ENABLE_NLS
 
   hard_LC_COLLATE = hard_locale (LC_COLLATE);
@@ -2292,7 +2347,6 @@ but lacks following character offset"));
 			else
 			  outfile = argv[++i];
 		      }
-		    close_stdout_set_file_name (outfile);
 		    goto outer;
 		  case 's':
 		    stable = 1;
@@ -2394,9 +2448,6 @@ but lacks following character offset"));
       files = &minus;
     }
 
-  if (sort_size == 0)
-    sort_size = default_sort_size ();
-
   if (checkonly)
     {
       if (nfiles > 1)
@@ -2409,89 +2460,16 @@ but lacks following character offset"));
       exit (check (files, nfiles) == 0 ? EXIT_SUCCESS : SORT_OUT_OF_ORDER);
     }
 
-  if (!STREQ (outfile, "-"))
-    {
-      struct stat outstat;
-      if (stat (outfile, &outstat) == 0)
-	{
-	  /* FIXME: warn about this */
-	  /* The following code prevents a race condition when
-	     people use the brain dead shell programming idiom:
-		  cat file | sort -o file
-	     This feature is provided for historical compatibility,
-	     but we strongly discourage ever relying on this in
-	     new shell programs. */
-
-	  /* Temporarily copy each input file that might be another name
-	     for the output file.  When in doubt (e.g. a pipe), copy.  */
-	  for (i = 0; i < nfiles; ++i)
-	    {
-	      char buf[8192];
-	      FILE *in_fp;
-	      FILE *out_fp;
-	      int cc;
-
-	      if (S_ISREG (outstat.st_mode) && !STREQ (outfile, files[i]))
-		{
-		  struct stat instat;
-		  if ((STREQ (files[i], "-")
-		       ? fstat (STDIN_FILENO, &instat)
-		       : stat (files[i], &instat)) != 0)
-		    die (files[i]);
-		  if (S_ISREG (instat.st_mode) && !SAME_INODE (instat, outstat))
-		    {
-		      /* We know the files are distinct.  */
-		      continue;
-		    }
-		}
-
-	      in_fp = xfopen (files[i], "r");
-	      tmp = create_temp_file (&out_fp);
-	      /* FIXME: maybe use copy.c(copy) here. */
-	      while ((cc = fread (buf, 1, sizeof buf, in_fp)) > 0)
-		write_bytes (buf, cc, out_fp, tmp);
-	      if (ferror (in_fp))
-		die (files[i]);
-	      xfclose (out_fp, tmp);
-	      xfclose (in_fp, files[i]);
-	      files[i] = tmp;
-	    }
-	  ofp = xfopen (outfile, "w");
-	}
-      else
-	{
-	  /* A non-`-' outfile was specified, but the file doesn't yet exist.
-	     Before opening it for writing (thus creating it), make sure all
-	     of the input files exist.  Otherwise, creating the output file
-	     could create an otherwise missing input file, making sort succeed
-	     when it should fail.  */
-	  for (i = 0; i < nfiles; ++i)
-	    {
-	      struct stat sb;
-	      if (STREQ (files[i], "-"))
-		continue;
-	      if (stat (files[i], &sb))
-		die (files[i]);
-	    }
-
-	  ofp = xfopen (outfile, "w");
-	}
-    }
-  else
-    {
-      ofp = stdout;
-    }
-
   if (mergeonly)
-    merge (files, nfiles, ofp, outfile);
+    {
+      int max_merge = first_same_file (files, MIN (nfiles, NMERGE), outfile);
+      merge (files, nfiles, max_merge, outfile);
+    }
   else
-    sort (files, nfiles, ofp, outfile);
-
-  if (fclose (ofp) != 0)
-    die (outfile);
+    sort (files, nfiles, outfile);
 
   if (have_read_stdin && fclose (stdin) == EOF)
-    die ("-");
+    die (_("close failed"), "-");
 
   cleanup ();
   exit (EXIT_SUCCESS);
