@@ -274,7 +274,8 @@ print the last N items in the file.  N may have a multiplier suffix:\n\
 b for 512, k for 1024, m for 1048576 (1 Meg).  A first OPTION of -VALUE\n\
 or +VALUE is treated like -n VALUE or -n +VALUE unless VALUE has one of\n\
 the [bkm] suffix multipliers, in which case it is treated like -c VALUE\n\
-or -c +VALUE.\n\
+or -c +VALUE.  Warning: a first option of +VALUE is obsolescent, and support\n\
+for it will be withdrawn.\n\
 \n\
 With --follow (-f), tail defaults to following the file descriptor, which\n\
 means that even if a tail'ed file is renamed, tail will continue to track\n\
